#pragma once

#include <boost/scoped_array.hpp>
#include <stdexcept>
#include <stdio.h>
#include <png.h>

#define PNG_BYTES_TO_CHECK 8

class pngfile {
private:
	FILE *fp;
	png_structp png;
	png_infop   info;
	png_uint_32 _width;
	png_uint_32 _height;

private:
	static void PNGAPI png_error_throw(png_structp png, png_const_charp msg)
	{
		throw std::runtime_error(msg);
	}

	bool is_valid_png(FILE *fp) {
		png_byte sig[PNG_BYTES_TO_CHECK];
		rewind(fp);
		if(fread(sig, 1, PNG_BYTES_TO_CHECK, fp) != PNG_BYTES_TO_CHECK)
			return false;
		if(!png_check_sig(sig, PNG_BYTES_TO_CHECK))
			return false;

		return true;
	}

public:
	pngfile() : fp(NULL), png(NULL), info(NULL) {
		png  = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, &png_error_throw, NULL);
		info = png_create_info_struct(png);
	}

	~pngfile()
	{
		png_destroy_read_struct(&png, &info, NULL);
		if(fp) fclose(fp);
	}


	bool open(const char *name)
	{
		fp = fopen(name, "rb");
		if(!fp) return false;

		if(!is_valid_png(fp)) {
			fclose(fp);
			fp = NULL;
			return false;
		}

		try {
			png_set_sig_bytes(png, PNG_BYTES_TO_CHECK);
			png_init_io(png, fp);
			png_read_info(png, info);

			int depth;
			int color;
			png_get_IHDR(png, info, &_width, &_height, &depth, &color, NULL, NULL, NULL);

			if(depth == 16) {
				png_set_strip_16(png);
			}

			switch(color) {
			case PNG_COLOR_TYPE_PALETTE:
				png_set_palette_to_rgb(png);
				png_set_filler(png, 255, PNG_FILLER_AFTER);
				break;
			case PNG_COLOR_TYPE_GRAY:
				png_set_gray_to_rgb(png);
				png_set_filler(png, 255, PNG_FILLER_AFTER);
				break;
			case PNG_COLOR_TYPE_GRAY_ALPHA:
				png_set_gray_to_rgb(png);
				break;
			case PNG_COLOR_TYPE_RGB:
				png_set_filler(png, 255, PNG_FILLER_AFTER);
				break;
			case PNG_COLOR_TYPE_RGB_ALPHA:
				break;
			}
			if (png_get_valid(png, info, PNG_INFO_tRNS)) {
				png_set_tRNS_to_alpha(png);
			}

			double gamma = .0f;
			if(png_get_gAMA(png, info, &gamma)) {
				png_set_gamma(png, 2.2f, gamma);
			}

			png_color_16 my_background = {0, 5, 5, 5, 0};
			png_color_16p image_background;

			if (png_get_bKGD(png, info, &image_background))
				png_set_background(png, image_background,
				  PNG_BACKGROUND_GAMMA_FILE, 1, 1.0);
			else
				png_set_background(png, &my_background,
				  PNG_BACKGROUND_GAMMA_SCREEN, 0, 1.0);

			png_set_bgr(png);

			png_read_update_info(png, info);

			return true;
		} catch(...) {
			fclose(fp);
			fp = NULL;
			return false;
		}
	}

	png_uint_32 height()
	{
		return fp ? _height : 0;
	}

	png_uint_32 width()
	{
		return fp ? _width : 0;
	}

	bool readall(png_bytep buf)
	{
		if(!fp) return false;

		try {
			boost::scoped_array<png_bytep> lines(new png_bytep[_height]);
			for (png_uint_32 i = 0; i < _height; ++i) {
				lines[_height - i - 1] = buf + _width * i * 4;
			}
			png_read_image(png, lines.get());
			png_read_end(png, info);

			return true;
		} catch (...) {
			return false;
		}
	}
};
